﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using System.Collections.Generic;
using System;

using BulletMonogameDemos.Demos;

namespace BulletMonogameDemos
{
    /// <summary>
    /// This is the main type for your game.
    /// </summary>
    public class Game1 : Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        DemoApplication current = null;
        SpriteFont font;
        public static Game1 Instance;
        Texture2D Button;
        Texture2D Over;
        List<Rectangle> button_array = new List<Rectangle>();
        int nDemos = 31;
        Rectangle rback = new Rectangle(10, 10, 256, 64);

        String[] button_names = new String[]
        {
            "Andrew",
            "Basic",
            "Box2D",
            //"BSP",
            "Bvh bug",
            //"Capsule heightfield",
            "Character",
            "Concave",
            "Concave raycast",
            "Constraint",
            "Corner",
            //"Meshes",
            "Forklift",
            //"Fracture",
            "Ghost object",
            "GImpact",
            "Internal edge",
            "Island stress",
            "Jenga",
            "Large mesh",
            "Motor",
            "Multi-world",
            "Pachinko",
            "Pyramid",
            "Ragdoll",
            "Simple compound",
            "Static level",
            "Terrain"
        };

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = 1920;
            graphics.PreferredBackBufferHeight = 1080;
            graphics.GraphicsProfile = GraphicsProfile.HiDef;
            graphics.SynchronizeWithVerticalRetrace = false;
            

            Content.RootDirectory = "Content";
            Instance = this;
            IsMouseVisible = true;
            nDemos = button_names.Length;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            int max_cnt = GraphicsDevice.Viewport.Width / 300;
            max_cnt = Math.Min(4, max_cnt);
            int xspace = (GraphicsDevice.Viewport.Width - (max_cnt * 256)) / (max_cnt + 1);
            int y_layers = ((nDemos / max_cnt) + 1);
            int yspace = (GraphicsDevice.Viewport.Height - (y_layers * 64)) / (y_layers + 1);

            int x = xspace;
            int y = yspace;
            int lc = 0;
            for (int i=0; i<nDemos; i++)
            {
                Rectangle r = new Rectangle(x, y, 256, 64);
                button_array.Add(r);
                x += 256 + xspace;
                lc++;
                if (lc == max_cnt)
                {
                    x = xspace;
                    y += 64 + yspace;
                    lc = 0;
                }
            }

            rback.X = GraphicsDevice.Viewport.Width - 300;
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            Button = Content.Load<Texture2D>("button");
            Over = Content.Load<Texture2D>("button_over");
            font = Content.Load<SpriteFont>("ButtonFont");
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// game-specific content.
        /// </summary>
        protected override void UnloadContent()
        {
            if (current != null)
                current.ShutdownDemo();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed || Keyboard.GetState().IsKeyDown(Keys.Escape))
                Exit();

            if (current != null)
            {
                current.Update(gameTime);
                MouseState ms = Mouse.GetState();
                if (ms.LeftButton == ButtonState.Pressed)
                {
                    if (rback.Contains(ms.X,ms.Y))
                    {
                        DemoApplication temp = current;
                        current = null;
                        temp.Cleanup();
                    }
                }
            }
            else
            {
                MouseState ms = Mouse.GetState();
                if (ms.LeftButton == ButtonState.Pressed)
                {
                    for (int i = 0; i < button_array.Count; i++)
                    {
                        if (button_array[i].Contains(ms.X, ms.Y))
                        {
                            DemoApplication temp = null;
                            String target = button_names[i];
                            switch (target)
                            {
                                case "Andrew":
                                    temp = new AndrewDemo();
                                    break;
                                case "Basic":
                                    temp = new BasicDemo();
                                    break;
                                case "Box2D":
                                    temp = new Box2DDemo();
                                    break;
                                case "BSP":
                                    break;
                                case "Bvh bug":
                                    temp = new BvhBugDemo();
                                    break;
                                case "Capsule heightfield":
                                    break;
                                case "Character":
                                    temp = new CharacterDemo();
                                    break;
                                case "Concave":
                                    temp = new ConcaveDemo();
                                    break;
                                case "Concave raycast":
                                    temp = new ConcaveRaycastDemo();
                                    break;
                                case "Constraint":
                                    temp = new ConstraintDemo();
                                    break;
                                case "Corner":
                                    temp = new CornerDemo();
                                    break;
                                case "Meshes":
                                    break;
                                case "Forklift":
                                    temp = new ForkLiftDemo();
                                    break;
                                case "Fracture":
                                    break;
                                case "Ghost object":
                                    temp = new GhostObjectDemo();
                                    break;
                                case "GImpact":
                                    temp = new GImpactTestDemo();
                                    break;
                                case "Internal edge":
                                    temp = new InternalEdgeDemo();
                                    break;
                                case "Island stress":
                                    temp = new IslandStressDemo();
                                    break;
                                case "Jenga":
                                    temp = new JengaDemo();
                                    break;
                                case "Large mesh":
                                    temp = new LargeMeshDemo();
                                    break;
                                case "Motor":
                                    temp = new MotorDemo();
                                    break;
                                case "Multi-world":
                                    temp = new MultiWorldDemo();
                                    break;
                                case "Pachinko":
                                    temp = new Pachinko();
                                    break;
                                case "Pyramid":
                                    temp = new PyramidDemo();
                                    break;
                                case "Ragdoll":
                                    temp = new RagDollDemo();
                                    break;
                                case "Simple compound":
                                    temp = new SimpleCompoundDemo();
                                    break;
                                case "Static level":
                                    temp = new StaticLevelDemo();
                                    break;
                                case "Terrain":
                                    temp = new TerrainDemo();
                                    break;
                            }
                            if (temp != null)
                            {
                                temp.InitializeDemo();
                                temp.LoadContent();
                                current = temp;
                            }
                        }
                    }
                }
            }
           
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            if (current != null)
            {
                current.Draw(gameTime);
                MouseState ms = Mouse.GetState();
                
                spriteBatch.Begin();

                if (rback.Contains(ms.X, ms.Y))
                    spriteBatch.Draw(Over, rback, Color.White);
                else
                    spriteBatch.Draw(Button, rback, Color.White);

                Vector2 pos = font.MeasureString("Back");
                pos.X = rback.X + 128 - (pos.X / 2);
                pos.Y = rback.Y + 32 - (pos.Y / 2);

                spriteBatch.DrawString(font, "Back", pos, Color.White);

                spriteBatch.End();
            }
            else
            {
                MouseState ms = Mouse.GetState();
                spriteBatch.Begin();
                for (int i = 0; i < nDemos; i++)
                {
                    if (button_array[i].Contains(ms.X, ms.Y))
                        spriteBatch.Draw(Over, button_array[i], Color.White);
                    else
                        spriteBatch.Draw(Button, button_array[i], Color.White);

                    Vector2 pos = font.MeasureString(button_names[i]);
                    pos.X = button_array[i].X + 128 - (pos.X / 2);
                    pos.Y = button_array[i].Y + 32 - (pos.Y / 2);

                    spriteBatch.DrawString(font, button_names[i], pos, Color.White);

                }
                spriteBatch.End();
            }
            

            base.Draw(gameTime);
        }
    }
}
